use crate::utils::{path, tests};
use crypto::digest::Digest;
use crypto::sha1::Sha1;
use std::ops::{Deref, DerefMut};
use std::path::PathBuf;
use std::sync::{LazyLock, Mutex, OnceLock};

// for tests only as it is mutable
static REPO_ROOT_DEV: LazyLock<Mutex<Vec<PathBuf>>> =
    LazyLock::new(|| Mutex::new(vec![PathBuf::new()]));

static REPO_ROOT: OnceLock<PathBuf> = OnceLock::new();

pub fn init(repo_root: &Option<PathBuf>) {
    if tests::is_var_setup() {
        REPO_ROOT_DEV.lock().unwrap()[0] = repo_root.clone().unwrap();
        return;
    }

    if let Some(repo_root) = repo_root {
        if REPO_ROOT.get().is_none() {
            let _ = REPO_ROOT.set(repo_root.clone());
        }
    } else {
        eprintln!("REPO_ROOT failed to initialize");
    }
}

pub fn get_repo_root() -> PathBuf {
    if tests::is_var_setup() {
        return REPO_ROOT_DEV.lock().unwrap()[0].clone();
    }

    match REPO_ROOT.get() {
        Some(path) => path.clone(),
        None => {
            panic!("fatal: 'REPO_ROOT' not set, you must initialize nsobject before using it!")
        }
    }
}

#[derive(Debug, PartialEq, Clone)]
pub struct ObjPath {
    path: PathBuf,
}

impl Deref for ObjPath {
    type Target = PathBuf;
    fn deref(&self) -> &PathBuf {
        &self.path
    }
}

impl DerefMut for ObjPath {
    fn deref_mut(&mut self) -> &mut PathBuf {
        &mut self.path
    }
}

pub fn to_obj_path(path: &PathBuf) -> ObjPath {
    ObjPath { path: path.clone() }
}

impl Into<ObjPath> for &PathBuf {
    fn into(self) -> ObjPath {
        ObjPath {
            path: path::to_repo_relative(self),
        }
    }
}

impl Into<ObjPath> for PathBuf {
    fn into(self) -> ObjPath {
        ObjPath {
            path: path::to_repo_relative(&self),
        }
    }
}

#[derive(Debug, PartialEq)]
pub struct NsObjPath {
    path: PathBuf,
}

impl Deref for NsObjPath {
    type Target = PathBuf;
    fn deref(&self) -> &PathBuf {
        &self.path
    }
}

impl DerefMut for NsObjPath {
    fn deref_mut(&mut self) -> &mut PathBuf {
        &mut self.path
    }
}

impl From<&str> for NsObjPath {
    fn from(hash: &str) -> Self {
        let (dir, res) = hash.split_at(2);

        let mut ns_obj_path = get_repo_root();
        ns_obj_path.push(dir);
        ns_obj_path.push(res);

        NsObjPath { path: ns_obj_path }
    }
}

impl From<&ObjPath> for NsObjPath {
    fn from(obj_path: &ObjPath) -> Self {
        let mut hasher = Sha1::new();
        hasher.input_str(
            obj_path
                .to_str()
                .expect("Cannot contains non UTF-8 char in path"),
        );
        NsObjPath::from(hasher.result_str().as_str())
    }
}
